<?php

namespace Drupal\context_redirect\Plugin\ContextReaction;

use Drupal\context\ContextManager;
use Drupal\context\Reaction\Annotation\ContextReaction;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Drupal\Core\Cache\CacheableRedirectResponse;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Path\PathValidatorInterface;
use Drupal\path_alias\AliasManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Routing\RedirectDestinationInterface;
use Drupal\Core\Url;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Component\Utility\Html;
use Drupal\context\ContextInterface;
use Drupal\context\ContextReactionPluginBase;
use Drupal\context\Form\AjaxFormTrait;

/**
 * Defines a Context reaction that will redirect to a url.
 *
 * @ContextReaction(
 *   id = "redirect",
 *   label = @Translation("Url Redirect"),
 *   description = "Provides a reaction that will redirect to a given Url."
 * )
 */
class Redirect extends ContextReactionPluginBase implements ContainerFactoryPluginInterface {

  use AjaxFormTrait;

  /** @var AliasManagerInterface */
  protected $pathAliasManager;

  /** @var RequestStack */
  protected $requestStack;

  /** @var RedirectDestinationInterface */
  protected $redirectDestination;

  /** @var PathValidatorInterface */
  protected $pathValidator;

  /** @var ModuleHandlerInterface */
  protected $moduleHandler;

  /** @var ContextManager */
  protected $contextManager;

  /**
   * @param array $configuration
   *   The plugin configuration.
   * @param $pluginId
   *   The plugin id.
   * @param $pluginDefinition
   *   The plugin definition.
   * @param \Drupal\path_alias\AliasManagerInterface $pathAliasManager
   *   The alias manager service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   * @param \Drupal\Core\Routing\RedirectDestinationInterface $redirectDestination
   *   The redirect destination service.
   * @param \Drupal\Core\Path\PathValidatorInterface $pathValidator
   *   The path validator service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler service.
   * @param \Drupal\context\ContextManager $contextManager
   *   The context manager service.
   */
  function __construct(array $configuration, $pluginId, $pluginDefinition, AliasManagerInterface $pathAliasManager, RequestStack $requestStack, RedirectDestinationInterface $redirectDestination, PathValidatorInterface $pathValidator, ModuleHandlerInterface $moduleHandler, ContextManager $contextManager) {
    parent::__construct($configuration, $pluginId, $pluginDefinition);
    $this->pathAliasManager = $pathAliasManager;
    $this->requestStack = $requestStack;
    $this->redirectDestination = $redirectDestination;
    $this->pathValidator = $pathValidator;
    $this->moduleHandler = $moduleHandler;
    $this->contextManager = $contextManager;
  }

  /**
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   * @param array $configuration
   * @param $pluginId
   * @param $pluginDefinition
   * @return \Drupal\context_redirect\Plugin\ContextReaction\Redirect|static
   */
  public static function create(ContainerInterface $container, array $configuration, $pluginId, $pluginDefinition) {
    return new static(
      $configuration,
      $pluginId,
      $pluginDefinition,
      $container->get('path_alias.manager'),
      $container->get('request_stack'),
      $container->get('redirect.destination'),
      $container->get('path.validator'),
      $container->get('module_handler'),
      $container->get('context.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'context_id' => NULL,
      'redirect_path' => NULL,
      'ignore_validation' => 0,
      'enable_message' => 0,
      'redirect_message' => NULL,
      'qsa' => 0,
      'redirect_destination' => 0,
      'redirect_type' => 302,
      'admin_key' => NULL,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->defaultConfiguration();
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function summary() {
    return $this->t('Redirect to a path on the site or an external URL.');
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state, ContextInterface $context = NULL) {
    $config = $this->getConfiguration();

    $form['context_id'] = [
      '#type' => 'value',
      '#value' => $context->id(),
    ];

    $form['redirect_path'] = [
      '#title' => $this->t('Redirect URL or path'),
      '#type' => 'textfield',
      '#default_value' => $config['redirect_path'] ?? '',
      '#description' => $this->t('Provide a valid internal URL (ie. node/1) or path (ie. company/about-us) or an external URL (ie. http://drupal.org). You may also use @front to redirect to the front page.', array('@front' => '<front>')),
    ];

    $form['ignore_validation'] = [
      '#title' => $this->t('Ignore path access validation'),
      '#type' => 'checkbox',
      '#default_value' => $config['ignore_validation'] ?? 0,
      '#description' => $this->t('Ignore path validation check when redirecting.'),
    ];

    $form['enable_message'] = [
      '#title' => $this->t('Display redirect message'),
      '#type' => 'checkbox',
      '#default_value' => $config['enable_message'] ?? 0,
      '#description' => $this->t('When enabled, this will display a message to the user saying they were redirected.'),
    ];

    $form['redirect_message'] = [
      '#title' => $this->t('Redirect message'),
      '#type' => 'textfield',
      '#default_value' => $config['redirect_message'] ?? '',
      '#description' => $this->t('The message displayed to the user after they are redirected.'),
      '#states' => [
        'visible' => [
          ':input[name="reactions[redirect][enable_message]"]' => [ 'checked' => TRUE ],
        ],
      ],
    ];

    $form['qsa'] = [
      '#title' => $this->t('Attach query string'),
      '#type' => 'checkbox',
      '#default_value' => $config['qsa'] ?? 0,
      '#description' => $this->t('When enabled, append the original query string (GET values) to redirected URL.'),
    ];

    $form['redirect_destination'] = [
      '#title' => $this->t('Redirect with destination'),
      '#type' => 'checkbox',
      '#default_value' => $config['redirect_destination'] ?? 0,
      '#description' => $this->t('Redirect to url with a destination query string parameter back to the current page.'),
    ];

    $form['redirect_type'] = [
      '#title' => $this->t('Redirect type'),
      '#type' => 'select',
      '#default_value' => $config['redirect_type'] ?? 302,
      '#description' => $this->t('The type of redirect to issue. You can @link.', [
        '@link' => Link::fromTextAndUrl('review HTTP redirect codes here',
          Url::fromUri('http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.3'))->toString()
      ]),
      '#options' => [
        301 => '301: Moved Permanently',
        302 => '302: Found',
        303 => '303: See Other',
        307 => '307: Temporary Redirect',
      ],
    ];

    $form['admin_key'] = [
      '#title' => $this->t('Admin Key'),
      '#type' => 'textfield',
      '#default_value' => $config['admin_key'] ?? '',
      '#field_prefix' => Url::fromUri('internal:', ['query' => ['admin_key' => ''], 'absolute' => TRUE]),
      '#description' => $this->t('You can optionally provide an "admin key" which you can enter as a get variable in the page url to prevent the redirect triggering (eg. ?admin_key=no-redirect-please your admin key would be "no-redirect-please"). Leave blank to disable this.'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $this->setConfiguration($values);
  }

  /**
   * {@inheritdoc}
   */
  public function execute() {
    $config = $this->getConfiguration();

    if ($redirect_path = $config['redirect_path']) {
      $redirect_path = $this->pathAliasManager->getPathByAlias($redirect_path);
      list($url, $uri) = $this->getUrlAndUriFromPath($redirect_path);
      $query = $this->requestStack->getCurrentRequest()->query;
      $base_url = $this->requestStack->getCurrentRequest()->getBaseUrl();

      // Do nothing if this reaction is an invalid redirection.
      if ($this->validateRedirect($redirect_path, $config)) {
        // Display a message if required.
        if ($config['enable_message']
          // Don't display message if the url is absolute and not local.
          && !$url->isExternal() && !UrlHelper::externalIsLocal($redirect_path, $base_url))
        {
          $message = Html::escape($config['redirect_message']);
          // todo: DI the Messenger service
          \Drupal::messenger()->addMessage('@message', ['@message' => $message]);
        }

        // Start building an $options array for the RedirectResponse.
        $options = [];
        // Preserve the current query parameters if required.
        if ($config['qsa']) {
          $options['query'] = $query;
        }
        if ($config['redirect_destination'] == 1) {
          // Add the current destination or the current path as the destination.
          $options['query']['destination'] = $this->redirectDestination->get();
        }

        $redirectUrl = Url::fromUri($uri, $options)->toString();
        $http_response_code = $config['redirect_type'] ?? 302;
        if ($context = $this->getContext()) {
          $response = new CacheableRedirectResponse($redirectUrl, $http_response_code);
          // Add a CacheableDependency on each of the context conditions.
          foreach ($context->getConditions() as $condition) {
            $response->addCacheableDependency($condition);
          }
          return $response;
        } else {
          // If the context is missing then we return an uncacheable redirect.
          return new RedirectResponse($redirectUrl, $http_response_code);
        }
      }
    }
  }

  /**
   * @return ContextInterface|null
   */
  protected function getContext() {
    $config = $this->getConfiguration();
    if ($context_id = ($config['context_id'] ?? NULL)) {
      try {
        if ($activeContexts = $this->contextManager->getActiveContexts()) {
          foreach ($activeContexts as $context) {
            if ($context->id() == $context_id) {
              return $context;
            }
          }
        }
      } catch (\Exception $e) {
        // todo: DI the Log service
        \Drupal::logger('context_redirect')->error($e->getMessage());
      }
    }
    return NULL;
  }

  /**
   * Validate the redirection.
   *
   * Depending on the current page, the destination url and the context's
   * configuration we may not want to allow this redirection.
   *
   * @param string $path
   *   The candidate redirect path / url to be validated.
   * @param array $config
   *   The context redirect reaction configuration.
   *
   * @return boolean
   *   TRUE if the redirect should be allowed, FALSE otherwise.
   */
  public function validateRedirect($path, $config) {
    $valid_redirect = TRUE;
    // todo: DI the Route Match service
    $current_route = \Drupal::routeMatch();
    $routeName = $current_route->getRouteName();
    $query = $this->requestStack->getCurrentRequest()->query;

    // Get the local path for the url if it exists.
    $path = $this->pathAliasManager->getPathByAlias($path);
    if ($config['ignore_validation'] !== 1){
      // If the url is not a valid path do nothing.
      if (!$this->pathValidator->isValid($path)) {
        $valid_redirect = FALSE;
      }
    }
    list($url, $uri) = $this->getUrlAndUriFromPath($path);

    // Resolve aliases and/or relative paths to absolute system urls if possible
    // and attempt to detect obvious redirect loops.
    $normalized_url = Url::fromUri($uri, ['absolute' => TRUE])->toString();
    $normalized_current_url = Url::fromRoute($routeName, $current_route->getRawParameters()->all(), ['absolute' => TRUE])->toString();
    if ($normalized_url == $normalized_current_url) {
      $valid_redirect = FALSE;
    }

    // Don't perform redirections on the context configuration screens.
    if ($routeName == 'entity.context.edit_form') {
      $valid_redirect = FALSE;
    }

    // Admin key override.
    if ($admin_key = $config['admin_key']) {
      if (($query['admin_key'] ?? NULL) == $admin_key) {
        $valid_redirect = FALSE;
      }
    }

    // Allow other modules to invalidate the redirect.
    $this->moduleHandler->alter('context_redirect_validate_redirect', $valid_redirect, $config);
    return $valid_redirect;
  }

  /**
   * Helper to get a normalized url and uri for a path.
   *
   * @param string $path
   * @return array
   */
  protected function getUrlAndUriFromPath(string $path) {
    if (UrlHelper::isExternal($path)) {
      $uri = $path;
    } else {
      $uri = strpos($path, 'internal:') === 0 ? $path : 'internal:' . $path;
    }
    $uri = rtrim($uri, '/') . '/';
    $url = Url::fromUri($uri);
    return [$url, $uri];
  }

}
